/*
    Copyright (C) 2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    aint with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef CANVASITEMS_H
#define CANVASITEMS_H

#include <qptrlist.h>
#include "CanvasItem.h"
#include "SVGTextElementImpl.h"

namespace T2P
{
	class GlyphSet;
	class GlyphLayoutParams;
};

namespace KSVG
{

class KSVGCanvas;
class KSVGTextChunk;
class SVGMatrixImpl;
class SVGMarkerElementImpl;
class SVGClipPathElementImpl;
class SVGTextPositioningElementImpl;

#define CANVAS_CLASS(Prefix, Class, Postfix, Member) \
class Canvas##Class : public CanvasItem \
{ \
public: \
	Canvas##Class(Prefix##Class##Postfix *Member) : CanvasItem(), m_##Member(Member) { } \
	virtual ~Canvas##Class() { } \
	virtual SVGElementImpl *element() { return reinterpret_cast<SVGElementImpl *>(m_##Member); } \
protected: \
	Prefix##Class##Postfix *m_##Member; \
};

CANVAS_CLASS(SVG, ClipPath, ElementImpl, clipPath)
CANVAS_CLASS(SVG, Marker, ElementImpl, marker)

class CanvasText : public CanvasItem
{
public:
	CanvasText(SVGTextElementImpl *text);
	virtual ~CanvasText();

	KSVGTextChunk *createTextChunk(KSVGCanvas *canvas, const SVGMatrixImpl *screenCTM, int &curx, int &cury, int &endx, int &endy);
	virtual SVGElementImpl *element() const { return m_text; }

	virtual void renderCallback(SVGTextPositioningElementImpl *element, const SVGMatrixImpl *screenCTM, T2P::GlyphSet *glyph, T2P::GlyphLayoutParams *params, double anchor, double baseline) const = 0;
	void createGlyphs(KSVGTextChunk *textChunk, KSVGCanvas *canvas, const SVGMatrixImpl *screenCTM, int curx, int cury, int &endx, int &endy) const;

	virtual void addTextDecoration(SVGTextPositioningElementImpl *element, double x, double y, double w, double h) const = 0;

protected:
	SVGTextElementImpl *m_text;
};

};

#endif

// vim:ts=4:noet
