/*
    Copyright (C) 2001-2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    aint with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <qimage.h>

#include "SVGPaint.h"
#include "SVGRectImpl.h"
#include "SVGAngleImpl.h"
#include "SVGPaintImpl.h"
#include "SVGMatrixImpl.h"
#include "SVGUnitTypes.h"
#include "SVGHelperImpl.h"
#include "SVGDocumentImpl.h"
#include "SVGPointListImpl.h"
#include "SVGMarkerElement.h"
#include "SVGSVGElementImpl.h"
#include "SVGPathSegListImpl.h"
#include "SVGAnimatedRectImpl.h"
#include "SVGAnimatedAngleImpl.h"
#include "SVGMarkerElementImpl.h"
#include "SVGAnimatedLengthImpl.h"
#include "SVGPolygonElementImpl.h"
#include "SVGClipPathElementImpl.h"
#include "SVGPolylineElementImpl.h"
#include "SVGAnimatedLengthListImpl.h"
#include "SVGAnimatedEnumerationImpl.h"
#include "SVGPreserveAspectRatioImpl.h"
#include "SVGAnimatedPreserveAspectRatioImpl.h"

#include "Glyph.h"
#include "Converter.h"
#include "KSVGTextChunk.h"

#include "agg_rasterizer_scanline_aa.h"
#include "agg_bounding_rect.h"
#include "agg_ellipse.h"
#include "agg_gen_stroke.h"
#include "agg_pixfmt_rgba32_image.h"
#include "agg_pixfmt_rgb24_image.h"
#include "agg_span_solid_rgba8.h"
#include "agg_scanline_u8.h"
#include "agg_rounded_rect.h"
typedef int scanline_type;
#include "agg_renderer_scanline.h"

#include "AggCanvasItems.h"
#include "AggCanvas.h"


using namespace KSVG;

// agg2 helpers

agg::gen_stroke::line_cap_e toAggLineCap(PathStrokeCapType cap)
{
	if(cap == PATH_STROKE_CAP_BUTT)
		return agg::gen_stroke::butt_cap;
	else if(cap == PATH_STROKE_CAP_ROUND)
		return agg::gen_stroke::square_cap;
	else
		return agg::gen_stroke::round_cap;
}

template<class Source>
dash_stroke_simple<Source>::dash_stroke_simple(Source& src, KSVG::SVGStylableImpl *style) : m_d(src), m_s(src), m_ds(m_d)
{
	unsigned int dashLength = style->getDashArray() ? style->getDashArray()->baseVal()->numberOfItems() : 0;
	m_dashed = dashLength > 0;
	if(m_dashed)
	{
		// there are dashes to be rendered
		for(unsigned int i = 0; i < dashLength; i += 2)
		m_d.add_dash(style->getDashArray()->baseVal()->getItem(i)->value(),
		 			 style->getDashArray()->baseVal()->getItem(i + 1)->value());
		m_d.dash_start(style->getDashOffset()->baseVal()->value());
		m_ds.width(style->getStrokeWidth()->baseVal()->value());
		m_ds.line_join((agg::gen_stroke::line_join_e)style->getJoinStyle());
		m_ds.miter_limit(style->getStrokeMiterlimit());
		m_ds.line_cap(toAggLineCap(style->getCapStyle()));
	}
	else
	{
		m_s.width(style->getStrokeWidth()->baseVal()->value());
		m_s.line_join((agg::gen_stroke::line_join_e)style->getJoinStyle());
		m_s.miter_limit(style->getStrokeMiterlimit());
		m_s.line_cap(toAggLineCap(style->getCapStyle()));
	}
}

// #####

AggShape::AggShape(AggCanvas *c, SVGStylableImpl *style)
: m_curved(m_storage), m_curved_trans(m_curved, m_transform), m_curved_trans_clipped(m_curved_trans), m_curved_stroked(m_curved, style), m_curved_stroked_trans(m_curved_stroked, m_transform), m_curved_stroked_trans_clipped(m_curved_stroked_trans), m_canvas(c), m_style(style)
{
	m_context = NORMAL;
}

AggShape::~AggShape()
{
	freeSVPs();
}

QRect AggShape::bbox() const
{
	return m_bbox;
}

bool AggShape::fillContains(const QPoint &p)
{
	agg::affine_matrix transform = m_transform;
	agg::affine_matrix mtx(m_canvas->zoom(), 0, 0, m_canvas->zoom(), m_canvas->pan().x(), m_canvas->pan().y());
	m_transform *= mtx;
	agg::rasterizer_scanline_aa<> ras;
	ras.filling_rule(m_style->getFillRule() == RULE_EVENODD ? agg::fill_even_odd : agg::fill_non_zero);
	ras.add_path(m_curved_trans);
	bool b = ras.hit_test(p.x(), p.y());
	m_transform = transform;
	return b;
}

bool AggShape::strokeContains(const QPoint &p)
{
	agg::affine_matrix transform = m_transform;
	agg::affine_matrix mtx(m_canvas->zoom(), 0, 0, m_canvas->zoom(), m_canvas->pan().x(), m_canvas->pan().y());
	m_transform *= mtx;
	agg::rasterizer_scanline_aa<> ras;
	ras.add_path(m_curved_stroked_trans);
	bool b = ras.hit_test(p.x(), p.y());
	m_transform = transform;
	return b;
}

void AggShape::update(CanvasItemUpdate reason, int, int)
{
	if(reason == UPDATE_STYLE)
		m_canvas->invalidate(this, false);
	else if(reason == UPDATE_TRANSFORM)
	{
		freeSVPs();
		init();
		m_canvas->invalidate(this, true);
	}
	else if(reason == UPDATE_ZOOM)
		init();
	else if(reason == UPDATE_PAN)
		init();
	else if(reason == UPDATE_LINEWIDTH)
	{
		init();
		m_canvas->invalidate(this, true);
	}
}

void AggShape::draw(SVGShapeImpl *shape)
{
	if(!m_style->getVisible() || !m_style->getDisplay() || !shape->directRender())
		return;

	//if(!m_strokeSVP && (!m_fillSVP || !m_style->isFilled()))
	//	init();
	agg::rect cb = m_canvas->m_ren.rbuf().clip_box();

	m_curved_trans_clipped.clip_box(cb.x1, cb.y1, cb.x2, cb.y2);
	m_curved_stroked_trans_clipped.clip_box(cb.x1, cb.y1, cb.x2, cb.y2);

	// transform ( zoom?)
	agg::affine_matrix transform = m_transform;
	agg::affine_matrix mtx(m_canvas->zoom(), 0, 0, m_canvas->zoom(), m_canvas->pan().x(), m_canvas->pan().y());
	m_transform *= mtx;
	m_curved.approximation_scale(pow(m_transform.scale(), 0.75));

	m_canvas->m_ras.reset();
	QColor qcolor;
	if(m_style->getFillColor()->paintType() == SVG_PAINTTYPE_CURRENTCOLOR)
		qcolor = m_style->getColor()->rgbColor().color();
	else
		qcolor = m_style->getFillColor()->rgbColor().color();
	agg::rgba8 color(qcolor.red(), qcolor.green(), qcolor.blue());
	color.opacity(m_style->getFillOpacity() * m_style->getOpacity());
	m_canvas->m_ren.attribute(color);
	m_canvas->m_ras.filling_rule(m_style->getFillRule() == RULE_EVENODD ? agg::fill_even_odd : agg::fill_non_zero);
	m_canvas->m_ras.add_path(m_curved_trans_clipped);
	if(m_style->isFilled())
		m_canvas->m_ras.render(m_canvas->m_ren);
	if(m_style->isStroked())
	{
		m_canvas->m_ras.reset();
		QColor qcolor;
		if(m_style->getStrokeColor()->paintType() == SVG_PAINTTYPE_CURRENTCOLOR)
			qcolor = m_style->getColor()->rgbColor().color();
		else
			qcolor = m_style->getStrokeColor()->rgbColor().color();
		agg::rgba8 color(qcolor.red(), qcolor.green(), qcolor.blue());
		color.opacity(m_style->getStrokeOpacity() * m_style->getOpacity());
		m_canvas->m_ren.attribute(color);
		m_canvas->m_ras.add_path(m_curved_stroked_trans_clipped);
		m_canvas->m_ras.render(m_canvas->m_ren);
	}
	m_transform = transform;
}

void AggShape::calcSVPs(const SVGMatrixImpl *matrix)
{
	// transform
	m_transform = agg::affine_matrix(matrix->a(), matrix->b(), matrix->c(), matrix->d(), matrix->e(), matrix->f());
	//m_curved = curved(*m_storage);
	//m_curved_trans = curved_trans(*m_curved, mtx);

	double x1, y1, x2, y2;
	agg::bounding_rect(m_curved_trans, *this, 0, 1, &x1, &y1, &x2, &y2);
	m_bbox = QRect(int(x1), int(y1), int(x2 - x1), int(y2 - y1));

	//if(m_style->isStroked())
	//{
	//	m_curved_stroked = new curved_stroked(*m_curved);
	//	m_curved_stroked_trans = new curved_stroked_trans(*m_curved_stroked, mtx);
	//}
}

void AggShape::init(const SVGMatrixImpl *)
{
}

void AggShape::freeSVPs()
{
	m_storage.remove_all();
}

// #####

AggRectangle::AggRectangle(AggCanvas *c, SVGRectElementImpl *rect)
: AggShape(c, rect), m_rect(rect)
{
	init();
}

void AggRectangle::draw()
{
	// Spec: a value of zero disables rendering
	if(m_rect->width()->baseVal()->value() > 0 && m_rect->height()->baseVal()->value() > 0)
		AggShape::draw(m_rect);
}

void AggRectangle::init()
{
	init(m_rect->screenCTM());
}

void AggRectangle::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	double x = m_rect->x()->baseVal()->value();
	double y = m_rect->y()->baseVal()->value();
	double width = m_rect->width()->baseVal()->value();
	double height = m_rect->height()->baseVal()->value();
	double rx = m_rect->rx()->baseVal()->value();
	double ry = m_rect->ry()->baseVal()->value();

	// Spec: If there is no rx or ry specified, draw a normal rect
	if(rx == -1 && ry == -1)
	{
		m_storage.move_to(x, y);
		m_storage.line_to(x + width, y);
		m_storage.line_to(x + width, y + height);
		m_storage.line_to(x, y + height);
	}
	else
	{
		int i = 0;

		// Spec: If rx isn't specified, but ry, set rx to ry
		if(rx == -1)
			rx = ry;

		// Spec: If ry isn't specified, but rx, set ry to rx
		if(ry == -1)
			ry = rx;

		// Spec: If rx is greater than half of the width of the rectangle
		//       then set rx to half of the width
		if(rx > width / 2)
			rx = width / 2;

		// Spec: If ry is greater than half of the height of the rectangle
		//       then set ry to half of the height
		if(ry > height / 2)
			ry = height / 2;

		m_storage.move_to(x + rx, y);

		i++;
		m_storage.curve4(x + rx * (1 - 0.552), y, x, y + ry * (1 - 0.552), x, y + ry);
		i++;
		if(ry < height / 2)
		{
			m_storage.line_to(x, y + height - ry);
			i++;
		}
		m_storage.curve4(x, y + height - ry * (1 - 0.552), x + rx * (1 - 0.552), y + height, x + rx, y + height);
		i++;
		if(rx < width / 2)
		{
			m_storage.line_to(x + width - rx, y + height);
			i++;
		}
		m_storage.curve4(x + width - rx * (1 - 0.552), y + height, x + width, y + height - ry * (1 - 0.552), x + width, y + height - ry);
		i++;
		if(ry < height / 2)
		{
			m_storage.line_to(x + width, y + ry);
			i++;
		}
		m_storage.curve4(x + width, y + ry * (1 - 0.552), x + width - rx * (1 - 0.552), y, x + width - rx, y);
		i++;
		if(rx < width / 2)
		{
			m_storage.line_to(x + rx, y);
			i++;
		}
	 }
	 m_storage.close_polygon();
	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(vec, m_rect, screenCTM, &m_fillSVP);
}

// #####

AggEllipse::AggEllipse(AggCanvas *c, SVGEllipseElementImpl *ellipse)
: AggShape(c, ellipse), m_ellipse(ellipse)
{
	init();
}
	
void AggEllipse::draw()
{
	// Spec: dont render when rx and/or ry is zero
	if(m_ellipse->rx()->baseVal()->value() > 0 && m_ellipse->ry()->baseVal()->value() > 0)
		AggShape::draw(m_ellipse);
}

void AggEllipse::init()
{
	init(m_ellipse->screenCTM());
}

void AggEllipse::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	double rx = m_ellipse->rx()->baseVal()->value();
	double ry = m_ellipse->ry()->baseVal()->value();
	double cx = m_ellipse->cx()->baseVal()->value();
	double cy = m_ellipse->cy()->baseVal()->value();

	agg::ellipse ell(cx, cy, rx, ry, 100);
	ell.rewind(0);
	double x, y;
	unsigned int cmd;
	while((cmd = ell.vertex(&x, &y)) != agg::path_cmd_stop)
		m_storage.add_vertex(x, y, cmd);
	m_storage.close_polygon();

	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(vec2, m_ellipse, screenCTM, &m_fillSVP);
}

// #####

AggCircle::AggCircle(AggCanvas *c, SVGCircleElementImpl *circle)
: AggShape(c, circle), m_circle(circle)
{
	init();
}
	
void AggCircle::draw()
{
	// Spec: a value of zero disables rendering
	if(m_circle->r()->baseVal()->value() > 0)
		AggShape::draw(m_circle);
}

void AggCircle::init()
{
	init(m_circle->screenCTM());
}

void AggCircle::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	double r = m_circle->r()->baseVal()->value();
	double cx = m_circle->cx()->baseVal()->value();
	double cy = m_circle->cy()->baseVal()->value();

	agg::ellipse ell(cx, cy, r, r, 100);
	ell.rewind(0);
	double x, y;
	unsigned int cmd;
	while((cmd = ell.vertex(&x, &y)) != agg::path_cmd_stop)
		m_storage.add_vertex(x, y, cmd);
	m_storage.close_polygon();

	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(vec2, m_ellipse, screenCTM, &m_fillSVP);
}

// #####

AggLine::AggLine(AggCanvas *c, SVGLineElementImpl *line)
: AggShape(c, line), m_line(line)
{
	init();
}

AggLine::~AggLine()
{
}

void AggLine::draw()
{
	if(!m_style->getVisible() || !m_style->getDisplay() || !m_line->directRender())
		return;

	agg::renderer_scanline_solid<agg::pixfmt_rgb24> ren(*(m_canvas->buf()));
	agg::rasterizer_scanline_aa<> ras;

	// transform ( zoom?)
	agg::affine_matrix transform = m_transform;
	agg::affine_matrix mtx(m_canvas->zoom(), 0, 0, m_canvas->zoom(), m_canvas->pan().x(), m_canvas->pan().y());
	m_transform *= mtx;
	m_curved.approximation_scale(pow(m_transform.scale(), 0.75));

	if(m_style->isStroked())
	{
		ras.reset();
		QColor qcolor;
		if(m_style->getStrokeColor()->paintType() == SVG_PAINTTYPE_CURRENTCOLOR)
			qcolor = m_style->getColor()->rgbColor().color();
		else
			qcolor = m_style->getStrokeColor()->rgbColor().color();
		agg::rgba8 color(qcolor.red(), qcolor.green(), qcolor.blue());
		color.opacity(m_style->getStrokeOpacity() * m_style->getOpacity());
		ren.attribute(color);
		/*m_curved_stroked.width(m_style->getStrokeWidth()->baseVal()->value());
		m_curved_stroked.line_join((agg::gen_stroke::line_join_e)m_style->getJoinStyle());
		m_curved_stroked.miter_limit(m_style->getStrokeMiterlimit());
		m_curved_stroked.line_cap(toAggLineCap(m_style->getCapStyle()));*/
		ras.add_path(m_curved_stroked_trans);
		ras.render(ren);
	}
	m_transform = transform;
}

void AggLine::init()
{
	init(m_line->screenCTM());
}

void AggLine::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	m_storage.move_to(m_line->x1()->baseVal()->value(), m_line->y1()->baseVal()->value());
	m_storage.line_to(m_line->x2()->baseVal()->value(), m_line->y2()->baseVal()->value());

	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(polyline, m_poly, screenCTM, &m_fillSVP);
}

// #####
AggPoly::AggPoly(AggCanvas *c, SVGPolyElementImpl *poly)
: AggShape(c, poly), m_poly(poly)
{
}

AggPoly::~AggPoly()
{
}

void AggPoly::init()
{
	init(m_poly->screenCTM());
}

void AggPoly::draw()
{
	AggShape::draw(m_poly);
}

// #####
AggPolyline::AggPolyline(AggCanvas *c, SVGPolylineElementImpl *poly, int points)
: AggPoly(c, poly), m_points(points)
{
	AggPoly::init();
}

AggPolyline::~AggPolyline()
{
}

void AggPolyline::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	QPointArray pa(m_poly->points()->numberOfItems());

	for(unsigned int i = 0; i < m_poly->points()->numberOfItems(); ++i)
		pa.setPoint(i, (int) m_poly->points()->getItem(i)->x(), (int) m_poly->points()->getItem(i)->y());

	if(pa.isEmpty() || pa.point(0).x() == -1 || pa.point(0).y() == -1)
		return;

	if(m_points == -1)
		m_points = pa.count();

	m_storage.move_to(pa.point(0).x(), pa.point(0).y());
	int index;
	for(index = 1; index < m_points; index++)
	{
		QPoint point = pa.point(index);
		m_storage.line_to(point.x(), point.y());
	}

	if(m_poly->isFilled()) // if the polyline must be filled, inform libart that it should not be closed.
		m_storage.close_polygon();

	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(polyline, m_poly, screenCTM, &m_fillSVP);
}

// #####

AggPolygon::AggPolygon(AggCanvas *c, SVGPolygonElementImpl *poly)
: AggPoly(c, poly)
{
	AggPoly::init();
}

AggPolygon::~AggPolygon()
{
}

void AggPolygon::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;

	QPointArray pa(m_poly->points()->numberOfItems());

	for(unsigned int i = 0; i < m_poly->points()->numberOfItems(); ++i)
		pa.setPoint(i, (int) m_poly->points()->getItem(i)->x(), (int) m_poly->points()->getItem(i)->y());

	if(pa.isEmpty())
		return;

	m_storage.move_to(pa.point(0).x(), pa.point(0).y());

	unsigned int index;
	for(index = 1; index < pa.count(); index++)
	{
		QPoint point = pa.point(index);
		m_storage.line_to(point.x(), point.y());
	}

	m_storage.line_to(pa.point(0).x(), pa.point(0).y());
	m_storage.close_polygon();
	//if(m_context == NORMAL)
		calcSVPs(screenCTM);
	//else
	//	calcClipSVP(polygon, m_poly, screenCTM, &m_fillSVP);
}

// #####

AggPath::AggPath(AggCanvas *c, SVGPathElementImpl *path)
: AggShape(c, path), SVGPathParser(), m_path(path)
{
	init();
}

AggPath::~AggPath()
{
}

void AggPath::draw()
{
	AggShape::draw(m_path);
}

void AggPath::init()
{
	init(m_path->screenCTM());
}

void AggPath::init(const SVGMatrixImpl *screenCTM)
{
	if(m_storage.total_vertices() > 0) return;
	if(!m_path->getAttribute("d").string().isEmpty())
	{
		m_storage.add_new_path();
		parseSVG(m_path->getAttribute("d").string(), true);

		// TODO : handle filled paths that are not closed explicitly

		// There are pure-moveto paths which reference paint servers *bah*
		// Do NOT render them
		bool dontrender = m_storage.total_vertices() == 1 && agg::is_move_to((*m_storage.begin()).cmd);
		if(!dontrender && m_context == NORMAL)
			calcSVPs(screenCTM);
		//else
		//	calcClipSVP(ksvg_art_bez_path_to_vec(m_array.data(), 0.25), m_path, screenCTM, &m_fillSVP);
	}
}

void AggPath::svgMoveTo(double x1, double y1, bool, bool)
{
	m_storage.move_to(x1, y1);
}

void AggPath::svgLineTo(double x1, double y1, bool)
{
	m_storage.line_to(x1, y1);
}

void AggPath::svgCurveToCubic(double x1, double y1, double x2, double y2, double x3, double y3, bool)
{
	m_storage.curve4(x1, y1, x2, y2, x3, y3);
}

void AggPath::svgClosePath()
{
	m_storage.close_polygon();
}

// #####

AggImage::AggImage(AggCanvas *c, SVGImageElementImpl *image)
: m_canvas(c), m_image(image)
{
}

AggImage::~AggImage()
{
}

void AggImage::draw()
{
	//SVGMatrixImpl *screenCTM = m_image->getScreenCTM();

	//float xval = m_image->x()->baseVal()->value();
	//float yval = m_image->y()->baseVal()->value();
	//float iwidth = m_image->width()->baseVal()->value();
	//float iheight = m_image->height()->baseVal()->value();

	/*m_canvas->setDrawShape(m_image);
	if(m_image->image()->width() != iwidth || m_image->image()->height() != iheight)
	{
		QImage show = m_image->image()->smoothScale(int(iwidth), int(iheight), QImage::ScaleMin);
		m_canvas->drawImage(show, m_image, screenCTM, int(xval), int(yval));
	}
	else
		m_canvas->drawImage(*m_image->image(), m_image, screenCTM, (int) xval, (int) yval);*/
	/*agg::rendering_buffer source_buffer;
	QImage *img = m_image->image();
	source_buffer.attach(img->bits(), img->width(), img->height(), img->width() * 4);

        typedef agg::image_transform_attr<agg::null_distortions, agg::null_color_alpha, agg::null_gradient_alpha> attr_type;
	typedef agg::span_null_interpolator<attr_type> span_interpolator;
	typedef agg::pixfmt_rgba32_image_bilinear<attr_type> span_renderer;
	typedef agg::renderer_scanline<span_interpolator, span_renderer> renderer;
	agg::null_distortions    distortions;
	agg::null_color_alpha    color_alpha;
	agg::null_gradient_alpha gradient_alpha;
	agg::affine_matrix img_mtx;
	attr_type attr(source_buffer, img_mtx, distortions, color_alpha, gradient_alpha);

	//agg::renderer_scanline<agg::span_solid_rgba8, agg::pixfmt_rgb24> ri(*(m_canvas->buf()));
	renderer ri(*(m_canvas->buf()));
	ri.attribute(attr);
	agg::rasterizer_scanline_aa<> pf;

	agg::rounded_rect viewp(xval, yval, img->width(), img->height(), 10);
	pf.add_path(viewp);
	pf.render(ri);*/
}

void AggImage::init()
{
}

QRect AggImage::bbox() const
{
	QRect bbox(static_cast<int>(m_image->x()->baseVal()->value()),
			   static_cast<int>(m_image->y()->baseVal()->value()),
			   static_cast<int>(m_image->width()->baseVal()->value()),
			   static_cast<int>(m_image->height()->baseVal()->value()));


	return SVGHelperImpl::fromUserspace(m_image, bbox);
}

// #####

AggText::AggText(AggCanvas *c, SVGTextElementImpl *text)
: CanvasText(text), m_canvas(c)
{
	init();
	m_drawItems.setAutoDelete(true);
}

AggText::~AggText()
{
}

bool AggText::fillContains(const QPoint &p)
{
	QPtrListIterator<SVPElement> it(m_drawItems);

	SVPElement *fill = it.current();
	while(fill)
	{
		if(fill->svp)
		{
			agg::rasterizer_scanline_aa<> ras;
			ras.filling_rule(fill->element->getFillRule() == RULE_EVENODD ? agg::fill_even_odd : agg::fill_non_zero);
			ras.add_path(fill->svp->m_curvedStrokedTrans);
			if(ras.hit_test(p.x(), p.y()))
				return true;
		}

		fill = ++it;
	}

	return false;
}

bool AggText::strokeContains(const QPoint &p)
{
	QPtrListIterator<SVPElement> it(m_drawItems);

	SVPElement *stroke = it.current();
	while(stroke)
	{
		if(stroke->svp)
		{
			agg::rasterizer_scanline_aa<> ras;
			ras.filling_rule(stroke->element->getFillRule() == RULE_EVENODD ? agg::fill_even_odd : agg::fill_non_zero);
			ras.add_path(stroke->svp->m_curvedStrokedTrans);
			if(ras.hit_test(p.x(), p.y()))
				return true;
		}

		stroke = ++it;
	}

	return false;
}

QRect AggText::bbox() const
{
	QRect result, rect;

	QPtrListIterator<SVPElement> it(m_drawItems);

	SVPElement *elem = it.current();
	while(elem)
	{
		double x1, y1, x2, y2;
		if(elem && elem->svp)
			agg::bounding_rect(elem->svp->m_curvedTrans, *this, 0, 1, &x1, &y1, &x2, &y2);

		rect.setX(int(x1));
		rect.setY(int(y1));
		rect.setWidth(int(x2 - x1));
		rect.setHeight(int(y2 - y1));

		result = result.unite(rect);

		elem = ++it;
	}

	return result;
}

void AggText::update(CanvasItemUpdate reason, int param1, int param2)
{
	if(reason == UPDATE_STYLE)
		m_canvas->invalidate(this, false);
	else if(reason == UPDATE_TRANSFORM)
	{
		clearCurved();
		init();
		m_canvas->invalidate(this, true);
	}
	else if(reason == UPDATE_ZOOM)
	{
		clearCurved();
		init();
	}
	else if(reason == UPDATE_PAN)
	{
		QPtrListIterator<SVPElement> it(m_drawItems);

		SVPElement *svpelement = it.current();
		T2P::BezierPathAgg *bpath;
		while(svpelement)
		{
			bpath = svpelement->svp;
			agg::affine_matrix mtx(1, 0, 0, 1, param1, param2);
			bpath->m_transform *= mtx;

			svpelement = ++it;
		}
	}
	/*
	else if(reason == UPDATE_LINEWIDTH)
	{
	}*/
}

void AggText::draw()
{
	QPtrListIterator<SVPElement> it(m_drawItems);

	SVPElement *svpelement = it.current();
	T2P::BezierPathAgg *bpath;
	SVGTextPositioningElementImpl *text;

	while(svpelement)
	{
		bpath = svpelement->svp;
		text = svpelement->element;
		if(!text->getVisible() || !text->getDisplay() || !text->directRender())
			return;

		// transform ( zoom?)
		//agg::affine_matrix transform = bpath->m_transform;
		//agg::affine_matrix mtx;//(m_canvas->zoom(), 0, 0, m_canvas->zoom(), 0, 0);
		//bpath->m_transform *= mtx;
		bpath->m_curved.approximation_scale(pow(bpath->m_transform.scale(), 0.75));

		m_canvas->m_ras.reset();
		QColor qcolor;
		if(text->getFillColor()->paintType() == SVG_PAINTTYPE_CURRENTCOLOR)
			qcolor = text->getColor()->rgbColor().color();
		else
			qcolor = text->getFillColor()->rgbColor().color();
		agg::rgba8 color(qcolor.red(), qcolor.green(), qcolor.blue());
		color.opacity(text->getFillOpacity() * text->getOpacity());
		m_canvas->m_ren.attribute(color);
		m_canvas->m_ras.filling_rule(text->getFillRule() == RULE_EVENODD ? agg::fill_even_odd : agg::fill_non_zero);
		m_canvas->m_ras.add_path(bpath->m_curvedTrans);
		if(text->isFilled())
			m_canvas->m_ras.render(m_canvas->m_ren);
		if(text->isStroked())
		{
			m_canvas->m_ras.reset();
			QColor qcolor;
			if(text->getStrokeColor()->paintType() == SVG_PAINTTYPE_CURRENTCOLOR)
				qcolor = text->getColor()->rgbColor().color();
			else
				qcolor = text->getStrokeColor()->rgbColor().color();
			agg::rgba8 color(qcolor.red(), qcolor.green(), qcolor.blue());
			color.opacity(text->getStrokeOpacity() * text->getOpacity());
			m_canvas->m_ren.attribute(color);
			bpath->m_curvedStroked.width(text->getStrokeWidth()->baseVal()->value());
			bpath->m_curvedStroked.line_join((agg::gen_stroke::line_join_e)text->getJoinStyle());
			bpath->m_curvedStroked.miter_limit(text->getStrokeMiterlimit());
			bpath->m_curvedStroked.line_cap(toAggLineCap(text->getCapStyle()));
			m_canvas->m_ras.add_path(bpath->m_curvedStrokedTrans);
			m_canvas->m_ras.render(m_canvas->m_ren);
		}
		//bpath->m_transform = transform;

		svpelement = ++it;
	}
}

void AggText::init()
{
	init(m_text->screenCTM());
}

void AggText::renderCallback(SVGTextPositioningElementImpl *element, const SVGMatrixImpl *screenCTM, T2P::GlyphSet *glyph, T2P::GlyphLayoutParams *params, double anchor, double baseline) const
{
	for(unsigned int i = 0; i < glyph->glyphCount(); i++)
	{
		T2P::GlyphAffinePair *glyphAffine = glyph->set().at(i);
		T2P::BezierPathAgg *bpath = static_cast<const T2P::BezierPathAgg *>(glyphAffine->transformatedPath());

		// text-anchor/baseline-shift support
		if(!params->tb())
			bpath->m_transform *= agg::affine_matrix(screenCTM->a(), screenCTM->b(), screenCTM->c(), screenCTM->d(), screenCTM->e() - anchor, screenCTM->f() + baseline);
		else
			bpath->m_transform *= agg::affine_matrix(screenCTM->a(), screenCTM->b(), screenCTM->c(), screenCTM->d(), screenCTM->e() - baseline, screenCTM->f() - anchor);

		SVPElement *svpelement = new SVPElement();
		svpelement->svp = new T2P::BezierPathAgg(*bpath);
		svpelement->element = element;

		m_drawItems.append(svpelement);
	}
}

void AggText::init(const SVGMatrixImpl *screenCTM)
{
	int curx = 0, cury = 0, endx = 0, endy = 0;
	KSVGTextChunk *textChunk = CanvasText::createTextChunk(m_canvas, screenCTM, curx, cury, endx, endy);

	if(textChunk->count() > 0)
		CanvasText::createGlyphs(textChunk, m_canvas, screenCTM, curx, cury, endx, endy);

	delete textChunk;
}

void AggText::clearCurved()
{
	m_drawItems.clear();
	// TODO: Huh - nobody does anything with the *trans* objects?:
}

void AggText::addTextDecoration(SVGTextPositioningElementImpl *element, double x, double y, double width, double height) const
{
	if(element->isFilled() || element->isStroked())
	{
		// compute rect 
		T2P::BezierPathAgg *bpath = new T2P::BezierPathAgg();
		bpath->storage.move_to(x, y);
		bpath->storage.line_to(x + width, y);
		bpath->storage.line_to(x + width, y + height);
		bpath->storage.line_to(x, y + height);

		const SVGMatrixImpl *mat = m_text->screenCTM();
		bpath->m_transform = agg::affine_matrix(mat->a(), mat->b(), mat->c(), mat->d(), mat->e(), mat->f());

		SVPElement *svpelement = new SVPElement();
		svpelement->svp = bpath;
		svpelement->element = element;

		m_drawItems.append(svpelement);
	}
}

AggText::SVPElement::~SVPElement()
{
	delete svp;
}

// vim:ts=4:noet
