/*
    Copyright (C) 2003 Nikolas Zimmermann <wildfox@kde.org>
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef T2P_BEZIERPATH_AGG_H
#define T2P_BEZIERPATH_AGG_H

#include <agg_curves.h>
#include <agg_conv_curve.h>
#include <agg_conv_stroke.h>
#include <agg_path_storage.h>
#include <agg_conv_transform.h>

#include "BezierPath.h"

namespace T2P
{
	typedef agg::conv_curve<agg::path_storage> curved;
	typedef agg::conv_transform<curved> curved_trans;
	typedef agg::conv_stroke<curved> curved_stroked;
	typedef agg::conv_transform<curved_stroked> curved_stroked_trans;

	class BezierPathAgg : public BezierPath
	{
	public:
		BezierPathAgg() : BezierPath(),
		m_curved(storage), m_curvedTrans(m_curved, m_transform),
		m_curvedStroked(m_curved), m_curvedStrokedTrans(m_curvedStroked, m_transform)
		{
		}

		~BezierPathAgg() 
		{
		}

		BezierPathAgg(const BezierPathAgg &other) : BezierPath(), storage(other.storage),
		 m_curved(storage), m_curvedTrans(m_curved, m_transform),
		m_curvedStroked(m_curved), m_curvedStrokedTrans(m_curvedStroked, m_transform)
		{
			m_transform = other.m_transform;
		}

		void copy_from(const agg::path_storage &ps, double *affine)
		{
		        for(unsigned i = 0; i < ps.total_vertices(); i++)
			{
				double x, y;
				unsigned cmd = ps.vertex(i, &x, &y);
				storage.add_vertex(affine[4] + x * affine[0] + y * affine[2], affine[5] + x * affine[1] + y * affine[3], cmd);
			}
		}

		agg::path_storage storage;
		curved m_curved;
		curved_trans m_curvedTrans;
		curved_stroked m_curvedStroked;
		curved_stroked_trans m_curvedStrokedTrans;
		agg::affine_matrix m_transform;
	};
};

#endif

// vim:ts=4:noet
